package MicroLang;

import java.io.LineNumberReader;
import java.io.IOException;

public class MicroParser {
  private final LineNumberReader inFile;
  private MicroToken currentToken;
  private final MicroScanner lexAnalyser;

  public MicroParser(LineNumberReader input)
  throws IOException, MicroLexicalError, MicroSyntaxError,
     MicroSemanticError {
    inFile = input;
    lexAnalyser = new MicroScanner(inFile);
    parse();
  }

  private void parse() throws IOException, MicroLexicalError,
                              MicroSyntaxError, MicroSemanticError {

    System.out.println("starting to parse ... ");
    currentToken = lexAnalyser.getNextToken();
    while (currentToken.getTokenKind() ==MicroToken.WhiteSpace)
      currentToken = lexAnalyser.getNextToken();
    BlockTy ast = program();

    System.out.println("finished parse ... ");
    /* adi */
    ast.identify (null);
    ast.typCompat ();
  }


   private BlockTy program()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    BlockTy ast = ((BlockTy)block());
    match(MicroToken.EOTtoken);
    return ast;
  }
  private StatTy block()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    BlockTy ast;
    ast = blockBegin();
    ast = (blockTail(ast));
    match(MicroToken.SemiColon);
    return ast;
  }
  private BlockTy blockBegin()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    BlockTy ast = new BlockTy(null, null, null, null, null);
    match(MicroToken.BlockSym);
    ast.Id1 = matchkeep(MicroToken.Ident);
    match(MicroToken.Colon);
    ast.DefPart  = defSeq();
    ast.StatPart = statSeq();
    match(MicroToken.EndSym);
    return ast;
  }
  private BlockTy blockTail(BlockTy ast)
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    switch (currentToken.getTokenKind()) {
      case (MicroToken.Ident):
        ast.Id2 = matchkeep(MicroToken.Ident);
        break;
      case (MicroToken.SemiColon):
        break;
      default:
        throw new MicroSyntaxError("Identifier or Semicolon expected " +
                                   "in Line " + inFile.getLineNumber());
    }
    return ast;
  }
  private DefTy defSeq()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    switch (currentToken.getTokenKind()) {
      case (MicroToken.VarSym):
        DefTy def = def();
        def.NextDef = defSeq();
        return def;
      case (MicroToken.AsgSym):
      case (MicroToken.IfSym):
      case (MicroToken.ForSym):
      case (MicroToken.WhileSym):
      case (MicroToken.ReadSym):
      case (MicroToken.WriteSym):
      case (MicroToken.EndSym):
      case (MicroToken.Ident):
      /* adi */
      case (MicroToken.BlockSym):
        return null;
      default:
        throw new MicroSyntaxError("keyword DEF or start of statement " +
                                   "sequence expected in line " +
                                   inFile.getLineNumber());
    }
  }
  private DefTy def()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    DefTy defReturnValue = new DefTy();
    match(MicroToken.VarSym);
    defReturnValue.Id = matchkeep(MicroToken.Ident);
    match(MicroToken.Colon);
    defReturnValue.Type = typeName();
    match(MicroToken.Assign);
    defReturnValue.Init = expr();
    match(MicroToken.SemiColon);
    return defReturnValue;
  }
  private MicroToken typeName()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    return (matchkeep(MicroToken.Ident));
  }
  private ExprTy expr()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    ExprTy left  = boolExpr1();
    left = exprTail(left);

    return left;
  }

  private ExprTy exprTail(ExprTy expression)
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    switch (currentToken.getTokenKind()) {
      case (MicroToken.Equiv):
        MicroToken operator = matchkeep(MicroToken.Equiv);
        ExprTy left  = new BinaryExprTy(new MicroToken (MicroToken.Equiv, null),
                                        expression,boolExpr1());
        left = exprTail(left);

        return left;

      case (MicroToken.SemiColon):
      case (MicroToken.RightParenthesis):
      case (MicroToken.ThenSym):
      case (MicroToken.Dots):
      case (MicroToken.LoopSym):
        return expression;
      default:
        throw new MicroSyntaxError("'<=>' or Semicolon, right bracket, " +
                                   "then, '..', or loop expected in line "+
                                   inFile.getLineNumber());
    }
  }

  private ExprTy boolExpr1()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    ExprTy left  = boolExpr2();
    left = boolExpr1Tail(left);
    return left;
  }


  private ExprTy boolExpr1Tail(ExprTy left)
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    switch(currentToken.getTokenKind()) {
      case (MicroToken.Impl):
        match(MicroToken.Impl);
        return new BinaryExprTy(new MicroToken(MicroToken.Impl, null),left,boolExpr2());
      case (MicroToken.SemiColon):
      case (MicroToken.RightParenthesis):
      case (MicroToken.ThenSym):
      case (MicroToken.Dots):
      case (MicroToken.LoopSym):
      case (MicroToken.Equiv):
        return left;
      default:
        throw new MicroSyntaxError("'=>' or Semicolon, right bracket, " +
                                   "then, '..', loop, or '<=>' expected " +
                                   "in line " + inFile.getLineNumber());
    }
  }

  private ExprTy boolExpr2()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    ExprTy left  = boolExpr3();
    left = boolExpr2Tail(left);
    return left;

  }

  private ExprTy boolExpr2Tail(ExprTy left)
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    switch (currentToken.getTokenKind()) {
      case (MicroToken.OrSym):
        match(MicroToken.OrSym);
        ExprTy right = new BinaryExprTy(new MicroToken(MicroToken.OrSym, null),
                                        left, boolExpr3());
        right = boolExpr2Tail(right);
        return right;
      case (MicroToken.SemiColon):
      case (MicroToken.RightParenthesis):
      case (MicroToken.ThenSym):
      case (MicroToken.Dots):
      case (MicroToken.LoopSym):
      case (MicroToken.Impl):
      case (MicroToken.Equiv):
        return left;
      default:
        throw new MicroSyntaxError("OR, semicolon, right bracket, " +
                                   "then, '..', loop '=>', or '<=>' " +
                                   "expected in line " +
                                   inFile.getLineNumber());
    }
  }

  private ExprTy boolExpr3()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    ExprTy left  = boolExpr4();
    left = boolExpr3Tail(left);
    return left;
  }

  private ExprTy boolExpr3Tail(ExprTy left)
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    switch(currentToken.getTokenKind()) {
      case (MicroToken.AndSym):
        match(MicroToken.AndSym);
        ExprTy right  = new BinaryExprTy(new MicroToken(MicroToken.AndSym, null),left,boolExpr4());
        right = boolExpr3Tail(right);
        return right;
      case (MicroToken.SemiColon):
      case (MicroToken.RightParenthesis):
      case (MicroToken.ThenSym):
      case (MicroToken.Dots):
      case (MicroToken.LoopSym):
      case (MicroToken.Impl):
      case (MicroToken.Equiv):
      case (MicroToken.OrSym):
        return left;
      default:
        throw new MicroSyntaxError("and, semicolon, right bracket, then, " +
                                   "'..' loop, '=>', '<=>', or 'or' " +
                                   "expected in line " +
                                   inFile.getLineNumber());
    }
  }

  private ExprTy boolExpr4()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    ExprTy left  = aritExpr1();
    left = boolExpr4Tail(left);
    return left;
  }

  private ExprTy boolExpr4Tail(ExprTy left)
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    switch (currentToken.getTokenKind()) {
      case (MicroToken.Equal):
      case (MicroToken.NotEqual):
      case (MicroToken.Less):
      case (MicroToken.LessEqual):
      case (MicroToken.Greater):
      case (MicroToken.GreaterEqual):
        MicroToken RelOperator = relOperator();
        ExprTy right = aritExpr1();
        return new BinaryExprTy(RelOperator,left, right);
      case (MicroToken.SemiColon):
      case (MicroToken.RightParenthesis):
      case (MicroToken.ThenSym):
      case (MicroToken.Dots):
      case (MicroToken.LoopSym):
      case (MicroToken.Impl):
      case (MicroToken.Equiv):
      case (MicroToken.OrSym):
      case (MicroToken.AndSym):
        return left;
      default:
        throw new MicroSyntaxError("relational Operator or Semicolon, " +
                                   "right bracket, then, '..', loop, " +
                                   "'=>', '<=>', 'or', or 'and' expected " +
                                   "in line "+inFile.getLineNumber());
    }
  }

  private MicroToken relOperator()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    switch (currentToken.getTokenKind()) {
      case (MicroToken.Equal):
      case (MicroToken.NotEqual):
      case (MicroToken.Less):
      case (MicroToken.LessEqual):
      case (MicroToken.Greater):
      case (MicroToken.GreaterEqual):
        return matchkeep(currentToken.getTokenKind());
      default:
        throw new MicroSyntaxError("relational Operator expected in Line " +
                                   inFile.getLineNumber());
    }
  }

  private ExprTy aritExpr1()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    ExprTy left;
    ExprTy right;
    switch(currentToken.getTokenKind()) {
      case (MicroToken.Ident):
      case (MicroToken.IntLit):
      case (MicroToken.LeftParenthesis):
      case (MicroToken.NotSym):
        left  = aritExpr2();
        right = aritExpr1Tail(left);
        return right;
      case (MicroToken.Plus):
      case (MicroToken.Minus):
        MicroToken UnaryOperator = matchkeep(currentToken.getTokenKind());
        left  = aritExpr2();
        left = new UnaryExprTy(UnaryOperator, left);
        right = aritExpr1Tail(left);
        return right;
      default:
        throw new MicroSyntaxError("Identifier, integer literal, left " +
                                   "bracket not symbol, plus, or minus " +
                                   "expected in line "+
                                   inFile.getLineNumber());
    }
  }

  private ExprTy aritExpr1Tail(ExprTy expression)
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    switch (currentToken.getTokenKind()) {
      case (MicroToken.Plus):
      case (MicroToken.Minus):
        MicroToken AddOperator = addOp();
        ExprTy left  = new BinaryExprTy(AddOperator, expression, aritExpr2());
        ExprTy right = aritExpr1Tail(left);
        return right;
      case (MicroToken.SemiColon):
      case (MicroToken.RightParenthesis):
      case (MicroToken.ThenSym):
      case (MicroToken.Dots):
      case (MicroToken.LoopSym):
      case (MicroToken.Impl):
      case (MicroToken.Equiv):
      case (MicroToken.OrSym):
      case (MicroToken.AndSym):
      case (MicroToken.Equal):
      case (MicroToken.NotEqual):
      case (MicroToken.Less):
      case (MicroToken.LessEqual):
      case (MicroToken.Greater):
      case (MicroToken.GreaterEqual):
        return expression;
      default:
        throw new MicroSyntaxError("additional or relational operator, "+
                                   "semicolon, right bracket, then, '..', "+
                                   " loop '=>', '<=>', 'or', or 'and' " +
                                   "expected in line "+
                                   inFile.getLineNumber());
    }
  }

  private MicroToken addOp()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    switch (currentToken.getTokenKind()) {
      case (MicroToken.Plus):
        return matchkeep(MicroToken.Plus);
      case (MicroToken.Minus):
        return matchkeep(MicroToken.Minus);
      default:
        throw new MicroSyntaxError("additional Operator expected in Line "+
                                   inFile.getLineNumber());
    }
  }

  private ExprTy aritExpr2()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    ExprTy left  = aritExpr3();
    ExprTy right = aritExpr2Tail(left);
    return right;
  }

  private ExprTy aritExpr2Tail(ExprTy expression)
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    switch (currentToken.getTokenKind()) {
      case (MicroToken.Times):
      case (MicroToken.Div):
        MicroToken MulOperator = mulOp();
        ExprTy left = new BinaryExprTy(MulOperator,expression,aritExpr3());
        ExprTy right = aritExpr2Tail(left);
        return right;
      case (MicroToken.SemiColon):
      case (MicroToken.RightParenthesis):
      case (MicroToken.ThenSym):
      case (MicroToken.Dots):
      case (MicroToken.LoopSym):
      case (MicroToken.Impl):
      case (MicroToken.Equiv):
      case (MicroToken.OrSym):
      case (MicroToken.AndSym):
      case (MicroToken.Equal):
      case (MicroToken.NotEqual):
      case (MicroToken.Less):
      case (MicroToken.LessEqual):
      case (MicroToken.Greater):
      case (MicroToken.GreaterEqual):
      case (MicroToken.Plus):
      case (MicroToken.Minus):
        return expression;
      default:
        throw new MicroSyntaxError("arithmetic, relational operation, " +
                                   "semicolon, right bracket, then, " +
                                   "'..', loop, '=>', '<=>', 'or', or " +
                                   "'and' expected in line "+
                                   inFile.getLineNumber());
    }
  }

  private MicroToken mulOp()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    switch (currentToken.getTokenKind()) {
      case (MicroToken.Times):
      case (MicroToken.Div):
        return matchkeep(currentToken.getTokenKind());
      default:
        throw new MicroSyntaxError("multiplication operator expected in " +
                                   "line "+ inFile.getLineNumber());
    }
  }

  private ExprTy aritExpr3()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    ExprTy left  = prim();
    ExprTy right = aritExpr3Tail();
    if (right==null)
      return left;
    else {
      ((BinaryExprTy)right).LeftOperand=left;
      return right;
    }

  }

  private ExprTy aritExpr3Tail()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    switch (currentToken.getTokenKind()) {
      case (MicroToken.Power):
        match(MicroToken.Power);
        ExprTy left  = prim();
        ExprTy right = aritExpr3Tail();
        if (right==null)
          return new BinaryExprTy(new MicroToken(MicroToken.Power, null), null, left);
        else {
          ((BinaryExprTy)right).LeftOperand=left;
          return new BinaryExprTy(new MicroToken(MicroToken.Power, null), null, right);
        }
      case (MicroToken.SemiColon):
      case (MicroToken.RightParenthesis):
      case (MicroToken.ThenSym):
      case (MicroToken.Dots):
      case (MicroToken.LoopSym):
      case (MicroToken.Impl):
      case (MicroToken.Equiv):
      case (MicroToken.OrSym):
      case (MicroToken.AndSym):
      case (MicroToken.Equal):
      case (MicroToken.NotEqual):
      case (MicroToken.Less):
      case (MicroToken.LessEqual):
      case (MicroToken.Greater):
      case (MicroToken.GreaterEqual):
      case (MicroToken.Plus):
      case (MicroToken.Minus):
      case (MicroToken.Times):
      case (MicroToken.Div):
        return null;
      default:
        throw new MicroSyntaxError("'**', aritmetic or relational " +
                                   "operator, semicolon, right " +
                                   "bracket, then, '..', loop '=>', " +
                                   "'<=>', 'or', or 'and' expected in " +
                                   "line " + inFile.getLineNumber());
    }
  }

  private ExprTy prim()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    switch (currentToken.getTokenKind()) {
      case (MicroToken.Ident):
        return new IdTy(matchkeep(MicroToken.Ident));
      case (MicroToken.IntLit):
        return new IntConstTy(matchkeep(MicroToken.IntLit));
      case (MicroToken.LeftParenthesis):
        match(MicroToken.LeftParenthesis);
        ExprTy exprInParenthesis = expr();
        match(MicroToken.RightParenthesis);
        return exprInParenthesis;
      case (MicroToken.NotSym):
        match(MicroToken.NotSym);
        ExprTy primary = prim();
        return new UnaryExprTy(new MicroToken(MicroToken.NotSym, null), primary);
      default:
        throw new MicroSyntaxError("Identifier, integer literal, " +
                                   "left bracket, or 'not' " +
                                   "expected in Line " +
                                   inFile.getLineNumber());
    }
  }

  private StatTy statSeq()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    switch (currentToken.getTokenKind()) {
      case (MicroToken.AsgSym):
      case (MicroToken.IfSym):
      case (MicroToken.ForSym):
      case (MicroToken.WhileSym):
      case (MicroToken.BlockSym):
      case (MicroToken.ReadSym):
      case (MicroToken.WriteSym):
        StatTy actualStat = stat();
        actualStat.NextStat = statSeq();
        return actualStat;
      case (MicroToken.EndSym):
      case (MicroToken.ElseSym):
        return null;
      default:
        throw new MicroSyntaxError("Beginning of statement, 'end', or " +
                                   "'else' expected in line " +
                                   inFile.getLineNumber());
    }
  }

  private StatTy stat()
    throws IOException, MicroSyntaxError, MicroLexicalError
  {
    ExprTy BoolExpr;
    StatTy LoopBody;
    switch (currentToken.getTokenKind()) {
      case (MicroToken.AsgSym):
        match(MicroToken.AsgSym);
        IdTy ident = new IdTy(matchkeep(MicroToken.Ident));
        match(MicroToken.Assign);
        ExprTy InitialValue = expr();
        match(MicroToken.SemiColon);
        return new AsgTy(ident, InitialValue, null);

      case (MicroToken.IfSym):
        match(MicroToken.IfSym);
        BoolExpr = expr();
        match(MicroToken.ThenSym);
        StatTy ThenPart = statSeq();
        match(MicroToken.ElseSym);
        StatTy ElsePart = statSeq();
        match(MicroToken.EndSym);
        match(MicroToken.IfSym);
        return new IfTy(BoolExpr, ThenPart, ElsePart, null);

      case (MicroToken.ForSym):
        match(MicroToken.ForSym);
        MicroToken LoopIdent = matchkeep(MicroToken.Ident);
        match(MicroToken.InSym);
        ExprTy LowerBound = expr();
        match(MicroToken.Dots);
        ExprTy UpperBound = expr();
        match(MicroToken.LoopSym);
        LoopBody = statSeq();
        match(MicroToken.EndSym);
        match(MicroToken.ForSym);
        return new ForTy(LoopIdent, LowerBound, UpperBound, LoopBody, null);

      case (MicroToken.WhileSym):
        match(MicroToken.WhileSym);
        BoolExpr = expr();
        match(MicroToken.LoopSym);
        LoopBody = statSeq();
        match(MicroToken.EndSym);
        match(MicroToken.WhileSym);
        System.out.println();
        return new WhileTy(BoolExpr, LoopBody, null);

      case (MicroToken.BlockSym):
        return block();

      case (MicroToken.ReadSym):
        match(MicroToken.ReadSym);
        MicroToken ReadIdent = matchkeep(MicroToken.Ident);
        match(MicroToken.SemiColon);
        return new ReadTy(new IdTy(ReadIdent), null);

      case (MicroToken.WriteSym):
        match(MicroToken.WriteSym);
        ExprTy WriteExpr = expr();
        match(MicroToken.SemiColon);
        return new WriteTy(WriteExpr, null);

      default:
        throw new MicroSyntaxError("Beginning of Statement expected in " +
                                   "Line "+ inFile.getLineNumber());
    }
  }

  private void match(int tokenKind)
    throws IOException, MicroLexicalError, MicroSyntaxError
  {
    if (currentToken.getTokenKind()==tokenKind) {
      if (currentToken.getTokenKind()!=MicroToken.EOTtoken) {
        currentToken = lexAnalyser.getNextToken();
        while (currentToken.getTokenKind() == MicroToken.WhiteSpace)
          currentToken = lexAnalyser.getNextToken();
      }
    }
    else {
      throw new MicroSyntaxError("Wrong Token in Line " +
                                 inFile.getLineNumber());
    }
  }

  private MicroToken matchkeep(int tokenKind)
            throws IOException, MicroLexicalError, MicroSyntaxError
  {
    MicroToken keepToken = currentToken;
    if (currentToken.getTokenKind()==tokenKind) {
      currentToken = lexAnalyser.getNextToken();
      while (currentToken.getTokenKind() == MicroToken.WhiteSpace)
        currentToken = lexAnalyser.getNextToken();
    }
    else {
      throw new MicroSyntaxError("Wrong Token in Line "+inFile.getLineNumber());
    }
    return keepToken;
  }

}